"""Utility functions for managing object state."""

import inspect
import types

from tippo import Any, Mapping
import six

from .import_path import get_path, import_path

__all__ = ["get_state", "update_state", "reducer"]


def get_state(obj):
    # type: (Any) -> dict[str, Any]
    """
    Get dictionary with an object's attribute values.
    Works with regular and slotted objects.

    :param obj: Object instance or class.
    :return: State dictionary.
    :raises TypeError: Provided object has no state.
    """
    if hasattr(obj, "__dict__"):
        return dict(obj.__dict__)

    if not hasattr(type(obj), "__slots__"):
        error = "{} object has no state".format(repr(type(obj).__name__))
        raise TypeError(error)

    state = {}  # type: dict[str, Any]
    cls = type(obj)
    for base in inspect.getmro(cls):
        if base is object:
            continue

        for slot in getattr(base, "__slots__", ()):

            # Skip weak reference slot.
            if slot == "__weakref__":
                continue

            # Privatize slot if needed.
            if slot.startswith("__") and not slot.endswith("__"):
                slot = "_{}{}".format(cls.__name__.lstrip("_"), slot)

            # Skip if already has a value.
            if slot in state:
                continue

            # Try to get value using the member descriptor.
            try:
                state[slot] = base.__dict__[slot].__get__(obj, cls)
            except (KeyError, AttributeError):
                pass

    return state


def update_state(obj, state_update):
    # type: (Any, Mapping[str, Any]) -> None
    """
    Update attribute values for an object.
    Works with regular and slotted objects.

    :param obj: Object instance or class.
    :param state_update: Dictionary with state updates.
    :raises TypeError: Provided object has no state.
    """
    if hasattr(obj, "__dict__"):
        if isinstance(obj, type):
            for attribute, value in six.iteritems(state_update):
                type.__setattr__(obj, attribute, value)
        else:
            obj.__dict__.update(state_update)
        return
    if not hasattr(type(obj), "__slots__"):
        error = "{} object has no state".format(repr(type(obj).__name__))
        raise TypeError(error)

    remaining = set(state_update)  # type: set[str]
    cls = type(obj)
    for base in inspect.getmro(cls):
        if base is object:
            continue

        for slot, value in six.iteritems(state_update):

            # Skip if already set.
            if slot not in remaining:
                continue

            # Skip slot if not in this base.
            if slot not in base.__dict__:
                continue

            # Get member.
            member = base.__dict__[slot]
            if not isinstance(member, types.MemberDescriptorType):
                error = "'{}.{}' is not a slot".format(cls.__name__, slot)
                raise AttributeError(error)

            # Set value using the member descriptor.
            member.__set__(obj, state_update[slot])
            remaining.remove(slot)

        if not remaining:
            return

    if remaining:
        error = "could not find slot(s) {} in {}".format(
            ", ".join(repr(s) for s in sorted(remaining)), repr(cls.__name__)
        )
        raise AttributeError(error)


def _reducer(path, state):
    cls = import_path(path)
    self = cls.__new__(cls)
    update_state(self, state)
    return self


def reducer(self):
    """Reducer method that supports qualified name and slots for Python 2.7."""
    return _reducer, (get_path(type(self)), get_state(self))



"""Utility to get mro bases."""

import collections
import inspect

from tippo import GenericMeta, Type

__all__ = ["get_mro"]


def get_mro(cls):
    # type: (Type) -> tuple[Type, ...]
    """
    Get consistent mro amongst different python versions.
    This works even with generic classes.

    :param cls: Class.
    :return: Tuple of classes.
    """

    # Resolve generic class to its origin.
    generic = cls
    while hasattr(generic, "__origin__"):
        generic = getattr(generic, "__origin__")
        if generic is not None:
            cls = generic

    # Newer python versions.
    if GenericMeta is type:
        return tuple(inspect.getmro(cls))

    # Special logic to skip generic classes when GenericMeta is being used, for old python.
    mro = collections.deque()  # type: collections.deque[Type]
    for base in reversed(inspect.getmro(cls)):
        generic = base
        origin = None
        while hasattr(generic, "__origin__"):
            generic = getattr(generic, "__origin__")
            if generic is not None:
                origin = generic
        if origin in mro:
            continue
        mro.appendleft(base)
    return tuple(mro)



"""Value factoring with support for lazy dot paths."""

import six
from tippo import Any, Callable, Iterable, Mapping, TypeVar

from .import_path import import_path

__all__ = ["fabricate_value", "format_factory"]


T = TypeVar("T")

MISSING = object()


def fabricate_value(
    factory,  # type: str | Callable | None
    value=MISSING,  # type: Any
    args=None,  # type: Iterable | None
    kwargs=None,  # type: Mapping[str, Any] | None
    extra_paths=(),  # type: Iterable[str]
    builtin_paths=None,  # type: Iterable[str] | None
):
    # type: (...) -> Any
    """
    Pass value through the factory.

    :param factory: Factory, import path, or None.
    :param value: Input value.
    :param args: Factory arguments.
    :param kwargs: Factory keyword arguments.
    :param extra_paths: Extra module paths in fallback order.
    :param builtin_paths: Builtin module paths in fallback order.
    :return: Output value.
    :raises ValueError: Factory is None and no value was provided.
    :raises TypeError: Invalid factory type.
    """
    if isinstance(factory, six.string_types):
        factory = import_path(factory, extra_paths=extra_paths, builtin_paths=builtin_paths)
    if factory is None:
        if value is MISSING:
            error = "no value and no factory provided"
            raise ValueError(error)
        return value
    if not callable(factory):
        error = "{!r} object is not a valid callable factory".format(type(factory).__name__)
        raise TypeError(error)
    if value is MISSING:
        return factory(*args or (), **kwargs or {})
    else:
        return factory(value, *args or (), **kwargs or {})


def format_factory(factory):
    # type: (str | Callable[..., T] | None) -> str | Callable[..., T] | None
    """
    Format and check factory.

    :param factory: Factory.
    :return: Checked factory.
    """
    if factory is not None and not isinstance(factory, six.string_types) and not callable(factory):
        error = "invalid factory type {!r}, expected None, a string, or a callable".format(type(factory).__name__)
        raise TypeError(error)
    return factory



import functools
import sys

import pytest  # noqa

from basicco.context_vars import ContextVar, Context, Token, copy_context


_var = ContextVar("_var")  # type: ContextVar[str]
_default_var = ContextVar("_default_var", default="default")  # type: ContextVar[str]


def _isolated_context(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        ctx = Context()
        return ctx.run(func, *args, **kwargs)

    return wrapper


def test_compat():
    if sys.version_info[0:2] >= (3, 7):
        import contextvars  # noqa

        assert ContextVar is contextvars.ContextVar
        assert Context is contextvars.Context
        assert Token is contextvars.Token
        assert copy_context is contextvars.copy_context
    else:
        with pytest.raises(ImportError):
            import contextvars  # noqa


def test_var():
    with pytest.raises(LookupError):
        _var.get()
    assert _var.get(None) is None

    token = _var.set("foo")
    assert _var.get() == "foo"
    _var.reset(token)

    with pytest.raises(LookupError):
        _var.get()  # type: ignore
    assert _var.get(None) is None


def test_default_var():
    assert _default_var.get() == "default"
    assert _var.get(None) is None

    token = _default_var.set("foo")
    assert _default_var.get() == "foo"
    _default_var.reset(token)

    assert _default_var.get() == "default"
    assert _var.get(None) is None


def test_context_var_new():
    with pytest.raises(TypeError):
        ContextVar()  # type: ignore

    exp = "must be a str"
    with pytest.raises(TypeError) as exc:
        ContextVar(1)  # type: ignore
    assert exp in str(exc)

    c = ContextVar("a")
    assert hash(c) != hash("a")


@_isolated_context
def test_context_var_repr_1():
    c = ContextVar("a")
    assert "a" in repr(c)

    c = ContextVar("a", default=123)
    assert "123" in repr(c)

    lst = []
    c = ContextVar("a", default=lst)
    lst.append(c)
    assert "..." in repr(c)
    assert "..." in repr(lst)

    t = c.set(1)
    assert repr(c) in repr(t)

    assert " used " not in repr(t)
    c.reset(t)
    assert " used " in repr(t)


def test_context_subclassing_1():
    exp = "not an acceptable base type"
    with pytest.raises(TypeError) as exc:

        class MyContextVar(ContextVar):  # type: ignore
            pass

        assert not MyContextVar
    assert exp in str(exc)

    exp = "not an acceptable base type"
    with pytest.raises(TypeError) as exc:

        class MyContext(Context):  # type: ignore
            pass

        assert not MyContext
    assert exp in str(exc)

    exp = "not an acceptable base type"
    with pytest.raises(TypeError) as exc:

        class MyToken(Token):  # type: ignore
            pass

        assert not MyToken
    assert exp in str(exc)


def test_context_new_1():
    with pytest.raises(TypeError):
        Context(1)  # type: ignore
    with pytest.raises(TypeError):
        Context(1, a=1)  # type: ignore
    with pytest.raises(TypeError):
        Context(a=1)  # type: ignore
    Context(**{})  # type: ignore


def test_context_type_errors_1():
    ctx = Context()

    exp = "ContextVar key was expected"
    with pytest.raises(TypeError) as exc:
        _ = ctx[1]  # type: ignore
    assert exp in str(exc)

    exp = "ContextVar key was expected"
    with pytest.raises(TypeError) as exc:
        _ = 1 in ctx
    assert exp in str(exc)

    exp = "ContextVar key was expected"
    with pytest.raises(TypeError) as exc:
        _ = ctx.get(1)  # type: ignore
    assert exp in str(exc)


def test_context_get_context_1():
    ctx = copy_context()
    assert isinstance(ctx, Context)


def test_context_run_1():
    ctx = Context()

    with pytest.raises(TypeError):
        ctx.run()  # type: ignore


def test_context_run_2():
    ctx = Context()

    def func(*args, **kwargs):
        kwargs["spam"] = "foo"
        args += ("bar",)
        return args, kwargs

    for f in (func, functools.partial(func)):
        assert ctx.run(f) == (("bar",), {"spam": "foo"})
        assert ctx.run(f, 1) == ((1, "bar"), {"spam": "foo"})

        assert ctx.run(f, a=2) == (("bar",), {"a": 2, "spam": "foo"})

        assert ctx.run(f, 11, a=2) == ((11, "bar"), {"a": 2, "spam": "foo"})

        a = {}
        assert ctx.run(f, 11, **a) == ((11, "bar"), {"spam": "foo"})
        assert a == {}


def test_context_run_3():
    ctx = Context()

    def func(*args, **kwargs):  # noqa
        _ = 1 / 0

    with pytest.raises(ZeroDivisionError):
        ctx.run(func)

    with pytest.raises(ZeroDivisionError):
        ctx.run(func, 1, 2)

    with pytest.raises(ZeroDivisionError):
        ctx.run(func, 1, 2, a=123)


@_isolated_context
def test_context_run_4():
    ctx1 = Context()
    ctx2 = Context()
    var = ContextVar("var")

    def func2():
        assert var.get(None) is None

    def func1():
        assert var.get(None) is None
        var.set("spam")
        ctx2.run(func2)
        assert var.get(None) == "spam"

        cur = copy_context()
        assert len(cur) == 1
        assert cur[var] == "spam"
        return cur

    returned_ctx = ctx1.run(func1)
    assert ctx1 == returned_ctx
    assert returned_ctx[var] == "spam"
    assert var in returned_ctx


def test_context_run_5():
    ctx = Context()
    var = ContextVar("var")

    def func():
        assert var.get(None) is None
        var.set("spam")
        _ = 1 / 0

    with pytest.raises(ZeroDivisionError):
        ctx.run(func)

    assert var.get(None) is None


def test_context_run_6():
    ctx = Context()
    c = ContextVar("a", default=0)

    def fun():
        assert c.get() == 0
        assert ctx.get(c) is None

        c.set(42)
        assert c.get() == 42
        assert ctx.get(c) == 42

    ctx.run(fun)


def test_context_run_7():
    ctx = Context()

    def fun():
        exp = "is already entered"
        with pytest.raises(RuntimeError) as exc:
            ctx.run(fun)
        assert exp in str(exc)

    ctx.run(fun)


@_isolated_context
def test_context_getset_1():
    c = ContextVar("c")
    with pytest.raises(LookupError):
        c.get()

    assert c.get(None) is None

    t0 = c.set(42)
    assert c.get() == 42
    assert c.get(None) == 42
    assert t0.old_value is t0.MISSING
    assert t0.old_value is Token.MISSING
    assert t0.var is c

    t = c.set("spam")
    assert c.get() == "spam"
    assert c.get(None) == "spam"
    assert t.old_value == 42
    c.reset(t)

    assert c.get() == 42
    assert c.get(None) == 42

    c.set("spam2")
    exp = "has already been used"
    with pytest.raises(RuntimeError) as exc:
        c.reset(t)
    assert exp in str(exc)

    assert c.get() == "spam2"

    ctx1 = copy_context()
    assert c in ctx1

    c.reset(t0)
    exp = "has already been used"
    with pytest.raises(RuntimeError) as exc:
        c.reset(t0)
    assert exp in str(exc)

    assert c.get(None) is None

    assert c in ctx1
    assert ctx1[c] == "spam2"
    assert ctx1.get(c, "aa") == "spam2"
    assert len(ctx1) == 1
    assert list(ctx1.items()) == [(c, "spam2")]
    assert list(ctx1.values()) == ["spam2"]
    assert list(ctx1.keys()) == [c]
    assert list(ctx1) == [c]

    ctx2 = copy_context()
    assert c not in ctx2
    with pytest.raises(KeyError):
        _ = ctx2[c]
    assert ctx2.get(c, "aa") == "aa"
    assert len(ctx2) == 0
    assert list(ctx2) == []


@_isolated_context
def test_context_getset_2():
    v1 = ContextVar("v1")
    v2 = ContextVar("v2")

    t1 = v1.set(42)
    exp = "by a different"
    with pytest.raises(ValueError) as exc:
        v2.reset(t1)
    assert exp in str(exc)


@_isolated_context
def test_context_getset_3():
    c = ContextVar("c", default=42)
    ctx = Context()

    def fun():
        assert c.get() == 42
        with pytest.raises(KeyError):
            _ = ctx[c]
        assert ctx.get(c) is None
        assert ctx.get(c, "spam") == "spam"
        assert c not in ctx
        assert list(ctx.keys()) == []

        t = c.set(1)
        assert list(ctx.keys()) == [c]
        assert ctx[c], 1

        c.reset(t)
        assert list(ctx.keys()) == []
        with pytest.raises(KeyError):
            _ = ctx[c]

    ctx.run(fun)


@_isolated_context
def test_context_getset_4():
    c = ContextVar("c", default=42)
    ctx = Context()

    tok = ctx.run(c.set, 1)

    exp = "different Context"
    with pytest.raises(ValueError) as exc:
        c.reset(tok)
    assert exp in str(exc)


@_isolated_context
def test_context_getset_5():
    c = ContextVar("c", default=42)  # type: ContextVar[int | list]
    c.set([])

    def fun():
        c.set([])
        c.get().append(42)
        assert c.get() == [42]

    copy_context().run(fun)
    assert c.get() == []


def test_context_copy_1():
    ctx1 = Context()
    c = ContextVar("c", default=42)

    def ctx1_fun():
        c.set(10)

        ctx2 = ctx1.copy()
        assert ctx2[c] == 10

        c.set(20)
        assert ctx1[c] == 20
        assert ctx2[c] == 10

        ctx2.run(ctx2_fun)
        assert ctx1[c] == 20
        assert ctx2[c] == 30

    def ctx2_fun():
        assert c.get() == 10
        c.set(30)
        assert c.get() == 30

    ctx1.run(ctx1_fun)


if __name__ == "__main__":
    pytest.main()



import pytest  # noqa

from basicco.fabricate_value import fabricate_value, format_factory


def test_fabricate_value():
    assert fabricate_value(None, 3) == 3
    assert fabricate_value(str, 3) == "3"
    assert fabricate_value("str", 3) == "3"


def test_format_factory():
    assert format_factory(None) is None
    assert format_factory(str) is str
    assert format_factory("str") == "str"

    with pytest.raises(TypeError):
        format_factory(3)  # type: ignore


if __name__ == "__main__":
    pytest.main()



import abc
import itertools

import pytest  # noqa
import tippo

from basicco.type_checking import (
    TypeCheckError,
    assert_is_callable,
    assert_is_instance,
    assert_is_subclass,
    assert_is_iterable,
    is_iterable,
    format_types,
    type_names,
    import_types,
    is_instance,
    is_subclass,
)


class Cls(object):
    pass


class SubCls(Cls):
    pass


class Parent(object):
    class Child(object):
        pass


cls_path = __name__ + "." + Cls.__name__
subcls_path = __name__ + "." + SubCls.__name__
nested_cls_path = __name__ + "." + Parent.__name__ + "." + Parent.Child.__name__


def test_format_types():
    assert format_types(None) == (type(None),)
    assert format_types((None,)) == (type(None),)
    assert format_types(float) == (float,)
    assert format_types("itertools.chain") == ("itertools.chain",)
    with pytest.raises(TypeError):
        format_types(3)  # type: ignore


def test_type_names():
    assert type_names(None) == (type(None).__name__,)
    assert type_names((None,)) == (type(None).__name__,)
    assert type_names(("module.module.Cls", "abc.abstractmethod", abc.ABCMeta, itertools.chain)) == (
        "Cls",
        "abstractmethod",
        "ABCMeta",
        "chain",
    )


def test_import_types():
    assert import_types(None) == (type(None),)
    assert import_types((None,)) == (type(None),)
    assert import_types(("abc.abstractmethod", abc.ABCMeta, itertools.chain)) == (
        abc.abstractmethod,
        abc.ABCMeta,
        itertools.chain,
    )


def test_is_instance():
    assert is_instance(None, None)
    assert is_instance(None, (None,))
    assert not is_instance(0, None)
    assert not is_instance(0, (None,))

    assert is_instance(Cls(), cls_path) is True
    assert is_instance(Cls(), cls_path, subtypes=False) is True
    assert is_instance(SubCls(), cls_path) is True
    assert is_instance(SubCls(), cls_path, subtypes=False) is False
    assert is_instance(SubCls(), subcls_path, subtypes=False) is True

    assert is_instance(Cls(), (cls_path, Cls, int)) is True
    assert is_instance(Cls(), (cls_path, Cls, int), subtypes=False) is True
    assert is_instance(SubCls(), (cls_path, Cls, int)) is True
    assert is_instance(SubCls(), (cls_path, Cls, int), subtypes=False) is False
    assert is_instance(SubCls(), (subcls_path, SubCls, int), subtypes=False) is True

    assert is_instance(Cls(), (int, float)) is False
    assert is_instance(Cls(), (int, float), subtypes=False) is False
    assert is_instance(SubCls(), (int, float)) is False
    assert is_instance(SubCls(), (int, float), subtypes=False) is False
    assert is_instance(SubCls(), (int, float), subtypes=False) is False

    assert is_instance(Cls(), ()) is False
    assert is_instance(SubCls(), ()) is False


def test_is_subclass():
    assert is_subclass(type(None), None)
    assert is_subclass(type(None), (None,))
    assert not is_subclass(int, None)
    assert not is_subclass(int, (None,))

    assert is_subclass(Cls, cls_path) is True
    assert is_subclass(Cls, cls_path, subtypes=False) is True
    assert is_subclass(SubCls, cls_path) is True
    assert is_subclass(SubCls, cls_path, subtypes=False) is False
    assert is_subclass(SubCls, subcls_path, subtypes=False) is True

    assert is_subclass(Cls, (cls_path, Cls, int)) is True
    assert is_subclass(Cls, (cls_path, Cls, int), subtypes=False) is True
    assert is_subclass(SubCls, (cls_path, Cls, int)) is True
    assert is_subclass(SubCls, (cls_path, Cls, int), subtypes=False) is False
    assert is_subclass(SubCls, (subcls_path, SubCls, int), subtypes=False) is True

    assert is_subclass(Cls, (int, float)) is False
    assert is_subclass(Cls, (int, float), subtypes=False) is False
    assert is_subclass(SubCls, (int, float)) is False
    assert is_subclass(SubCls, (int, float), subtypes=False) is False
    assert is_subclass(SubCls, (int, float), subtypes=False) is False

    assert is_subclass(Cls, ()) is False
    assert is_subclass(SubCls, ()) is False


def test_is_iterable():
    assert not is_iterable(3)
    assert not is_iterable(False)
    assert not is_iterable("foo")

    assert is_iterable("foo", include_strings=True)
    assert is_iterable([])
    assert is_iterable({})
    assert is_iterable(set())


def test_assert_is_instance():
    assert_is_instance(None, None)
    assert_is_instance(None, (None,))
    with pytest.raises(TypeCheckError):
        assert_is_instance(0, None)
        assert_is_instance(0, (None,))

    assert_is_instance(Cls(), cls_path)
    assert_is_instance(Cls(), cls_path, subtypes=False)
    assert_is_instance(SubCls(), cls_path)

    with pytest.raises(TypeCheckError):
        assert_is_instance(SubCls(), cls_path, subtypes=False)

    assert_is_instance(SubCls(), subcls_path, subtypes=False)

    assert_is_instance(Cls(), (cls_path, Cls, int))
    assert_is_instance(Cls(), (cls_path, Cls, int), subtypes=False)
    assert_is_instance(SubCls(), (cls_path, Cls, int))

    with pytest.raises(TypeCheckError):
        assert_is_instance(SubCls(), (cls_path, Cls, int), subtypes=False)

    assert_is_instance(SubCls(), (subcls_path, SubCls, int), subtypes=False)

    with pytest.raises(TypeCheckError):
        assert_is_instance(Cls(), (int, float))

    with pytest.raises(TypeCheckError):
        assert_is_instance(Cls(), (int, float), subtypes=False)

    with pytest.raises(TypeCheckError):
        assert_is_instance(SubCls(), (int, float))

    with pytest.raises(TypeCheckError):
        assert_is_instance(SubCls(), (int, float), subtypes=False)

    with pytest.raises(TypeCheckError):
        assert_is_instance(SubCls(), (int, float), subtypes=False)

    with pytest.raises(ValueError):
        assert_is_instance(Cls(), ())

    with pytest.raises(ValueError):
        assert_is_instance(SubCls(), ())


def test_assert_is_subclass():
    assert_is_subclass(type(None), None)
    assert_is_subclass(type(None), (None,))
    with pytest.raises(TypeCheckError):
        assert_is_subclass(int, None)
        assert_is_subclass(int, (None,))

    assert_is_subclass(Cls, cls_path)
    assert_is_subclass(Cls, cls_path, subtypes=False)
    assert_is_subclass(SubCls, cls_path)

    with pytest.raises(TypeCheckError):
        assert_is_subclass(SubCls, cls_path, subtypes=False)

    assert_is_subclass(SubCls, subcls_path, subtypes=False)

    assert_is_subclass(Cls, (cls_path, Cls, int))
    assert_is_subclass(Cls, (cls_path, Cls, int), subtypes=False)
    assert_is_subclass(SubCls, (cls_path, Cls, int))

    with pytest.raises(TypeCheckError):
        assert_is_subclass(SubCls, (cls_path, Cls, int), subtypes=False)

    assert_is_subclass(SubCls, (subcls_path, SubCls, int), subtypes=False)

    with pytest.raises(TypeCheckError):
        assert_is_subclass(Cls, (int, float))  # type: ignore

    with pytest.raises(TypeCheckError):
        assert_is_subclass(Cls, (int, float), subtypes=False)  # type: ignore

    with pytest.raises(TypeCheckError):
        assert_is_subclass(SubCls, (int, float))  # type: ignore

    with pytest.raises(TypeCheckError):
        assert_is_subclass(SubCls, (int, float), subtypes=False)  # type: ignore

    with pytest.raises(TypeCheckError):
        assert_is_subclass(SubCls, (int, float), subtypes=False)  # type: ignore

    with pytest.raises(ValueError):
        assert_is_subclass(Cls, ())

    with pytest.raises(ValueError):
        assert_is_subclass(SubCls, ())


def test_assert_is_callable():
    assert_is_callable(int)

    with pytest.raises(TypeCheckError):
        assert_is_callable(3)  # type: ignore


def test_assert_is_iterable():
    with pytest.raises(TypeCheckError):
        assert_is_iterable(3)  # type: ignore

    with pytest.raises(TypeCheckError):
        assert_is_iterable(False)  # type: ignore

    with pytest.raises(TypeCheckError):
        assert_is_iterable("foo")

    assert_is_iterable("foo", include_strings=True)

    assert_is_iterable([])
    assert_is_iterable({})
    assert_is_iterable(set())


def test_typing_type():
    assert is_instance(3.0, float) == isinstance(3.0, float)
    assert is_instance(float, tippo.Type[float]) == is_subclass(float, float)

    assert is_subclass(float, tippo.Type[tippo.Type[float]]) == is_subclass(float, type(float))
    assert is_subclass(float, tippo.Type[int]) == is_subclass(float, type)

    assert is_instance({"a": 1}, tippo.Mapping[str, int])
    assert is_instance({"a": 1}, tippo.Dict[str, int])
    assert is_instance(["a"], tippo.Iterable[str])
    assert is_instance(["a"], tippo.List[str])
    assert is_instance({"a"}, tippo.Iterable[str])
    assert is_instance({"a"}, tippo.Set[str])

    assert not is_instance({"a": 1}, tippo.Mapping[float, str])
    assert not is_instance({"a": 1}, tippo.Dict[float, str])
    assert not is_instance(["a"], tippo.Iterable[int])
    assert not is_instance(["a"], tippo.List[int])
    assert not is_instance({"a"}, tippo.Iterable[int])
    assert not is_instance({"a"}, tippo.Set[int])

    assert is_instance(dict, tippo.Type[tippo.Mapping])
    assert is_instance(dict, tippo.Type[tippo.Dict])
    assert is_instance(list, tippo.Type[tippo.Iterable])
    assert is_instance(list, tippo.Type[tippo.List])
    assert is_instance(set, tippo.Type[tippo.Iterable])
    assert is_instance(set, tippo.Type[tippo.Set])

    assert is_instance(dict, tippo.Type[tippo.Mapping[int, str]])
    assert is_instance(dict, tippo.Type[tippo.Dict[int, str]])
    assert is_instance(list, tippo.Type[tippo.Iterable[int]])
    assert is_instance(list, tippo.Type[tippo.List[int]])
    assert is_instance(set, tippo.Type[tippo.Iterable[int]])
    assert is_instance(set, tippo.Type[tippo.Set[int]])

    assert is_instance(3, tippo.Any)
    assert is_instance(True, tippo.Literal[True, False])
    assert is_instance(3, tippo.Union[int, str])
    assert is_instance(int, tippo.Type[int])
    assert is_instance((3, "3"), tippo.Tuple[int, str])
    assert is_instance((3, 4, 5), tippo.Tuple[int, ...])
    assert is_instance({"3": 3}, tippo.Mapping[str, int])
    assert is_instance([1, 2, 3], tippo.Iterable[int])

    assert not is_instance(None, tippo.Literal[True, False])
    assert not is_instance(3.4, tippo.Union[int, str])
    assert not is_instance(float, tippo.Type[int])
    assert not is_instance(("3", 3), tippo.Tuple[int, str])
    assert not is_instance(("3", "4", "5"), tippo.Tuple[int, ...])
    assert not is_instance({3: "3"}, tippo.Mapping[str, int])
    assert not is_instance([1, 2, 3], tippo.Iterable[str])

    assert is_instance(Parent.Child(), tippo.Optional[nested_cls_path])


if __name__ == "__main__":
    pytest.main()



import pytest  # noqa
from tippo import TypeVar, Generic

from basicco.get_mro import get_mro


T = TypeVar("T")


def test_generic_mro():
    T = TypeVar("T")  # type: ignore  # noqa

    class Base(Generic[T]):
        pass

    class SubBase(Base[T]):
        pass

    class SubSubBase(SubBase[T], Base[T]):
        pass

    assert get_mro(SubSubBase) == (SubSubBase, SubBase, Base, Generic, object)
    assert get_mro(SubSubBase[T]) == (SubSubBase, SubBase, Base, Generic, object)
    assert get_mro(SubSubBase[int]) == (SubSubBase, SubBase, Base, Generic, object)


if __name__ == "__main__":
    pytest.main()
